/*----- PROTECTED REGION ID(E621.cpp) ENABLED START -----*/
//=============================================================================
//
// file :        E621.cpp
//
// description : C++ source for the E621 class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               E621 are implemented in this file.
//
// project :     
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
// $Author:  $
//
// $Revision:  $
// $Date:  $
//
// $HeadURL:  $
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <E621.h>
#include <E621Class.h>
#include <iomanip>

#define MAX_DEVSTRING_LENGTH 255

/*----- PROTECTED REGION END -----*/	//	E621.cpp

/**
 *  E621 class description:
 *    
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name  |  Method name
//================================================================
//  State         |  Inherited (no method)
//  Status        |  Inherited (no method)
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  Firmware      |  Tango::DevString	Scalar
//  SerialNumber  |  Tango::DevString	Scalar
//  AverageTimes  |  Tango::DevUShort	Scalar
//================================================================

namespace E621_ns
{
/*----- PROTECTED REGION ID(E621::namespace_starting) ENABLED START -----*/

//	static initializations
template<typename F, typename T> void convert(const F &data, T &result)
{
	stringstream ss;
	ss << data;
	ss >> result;

	if (ss.fail()) {
		stringstream err_mesg;
		err_mesg << "Conversion fail of " << data;
		Tango::Except::throw_exception( "",
				err_mesg.str(),
				"convert()");
	}
}

template<typename F, typename T> void convert(const F &data, double &result)
{
	stringstream ss;
	ss << data;

	if (ss.str() == "Inf")
		result = INFINITY;
	else
		ss >> result;

	if (ss.fail()) {
		stringstream err_mesg;
		err_mesg << "Conversion fail of " << data;
		Tango::Except::throw_exception( "",
				err_mesg.str(),
				"convert()");
	}
}

template<typename F> void convert(const F &data, string &result)
{
	stringstream ss;
	ss << fixed;
	ss << data;
	result = ss.str();
	if (ss.fail()) {
		stringstream err_mesg;
		err_mesg << "Conversion fail of " << data;
		Tango::Except::throw_exception( "",
				err_mesg.str(),
				"convert()");
	}
}

/*----- PROTECTED REGION END -----*/	//	E621::namespace_starting

//--------------------------------------------------------
/**
 *	Method     : E621::E621()
 *	Description: Constructors for a Tango device
 *                implementing the classE621
 */
//--------------------------------------------------------
E621::E621(Tango::DeviceClass *cl, std::string &s)
 : USB2Client(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(E621::constructor_1) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	E621::constructor_1
}
//--------------------------------------------------------
E621::E621(Tango::DeviceClass *cl, const char *s)
 : USB2Client(cl, s)
{
	/*----- PROTECTED REGION ID(E621::constructor_2) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	E621::constructor_2
}
//--------------------------------------------------------
E621::E621(Tango::DeviceClass *cl, const char *s, const char *d)
 : USB2Client(cl, s, d)
{
	/*----- PROTECTED REGION ID(E621::constructor_3) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	E621::constructor_3
}
//--------------------------------------------------------
E621::~E621()
{
	delete_device();
}

//--------------------------------------------------------
/**
 *	Method     : E621::delete_device()
 *	Description: will be called at device destruction or at init command
 */
//--------------------------------------------------------
void E621::delete_device()
{
	DEBUG_STREAM << "E621::delete_device() " << device_name << std::endl;
	/*----- PROTECTED REGION ID(E621::delete_device) ENABLED START -----*/
	
	//	Delete device allocated objects
	if (attr_Firmware_read[0])
		delete[] attr_Firmware_read[0];
	if (attr_SerialNumber_read[0])
		delete[] attr_SerialNumber_read[0];

	/*----- PROTECTED REGION END -----*/	//	E621::delete_device
	delete[] attr_Firmware_read;
	delete[] attr_SerialNumber_read;
	delete[] attr_AverageTimes_read;

	if (Tango::Util::instance()->is_svr_shutting_down()==false  &&
		Tango::Util::instance()->is_device_restarting(device_name)==false &&
		Tango::Util::instance()->is_svr_starting()==false)
	{
		//	If not shutting down call delete device for inherited object
		USB2Client_ns::USB2Client::delete_device();
	}
}

//--------------------------------------------------------
/**
 *	Method     : E621::init_device()
 *	Description: will be called at device initialization.
 */
//--------------------------------------------------------
void E621::init_device()
{
	DEBUG_STREAM << "E621::init_device() create device " << device_name << std::endl;
	/*----- PROTECTED REGION ID(E621::init_device_before) ENABLED START -----*/
	
	//	Initialization before get_device_property() call
	mandatoryNotDefined = false;
	init_error.clear();
	
	/*----- PROTECTED REGION END -----*/	//	E621::init_device_before

	if (Tango::Util::instance()->is_svr_starting() == false  &&
		Tango::Util::instance()->is_device_restarting(device_name)==false)
	{
		//	If not starting up call init device for inherited object
		USB2Client_ns::USB2Client::init_device();
	}
	//	No device property to be read from database

	attr_Firmware_read = new Tango::DevString[1];
	attr_SerialNumber_read = new Tango::DevString[1];
	attr_AverageTimes_read = new Tango::DevUShort[1];
	//	No longer if mandatory property not set.
	if (mandatoryNotDefined)
		return;

	/*----- PROTECTED REGION ID(E621::init_device) ENABLED START -----*/
	
	//	Initialize device
	string rep;
	try {
		attr_Firmware_read[0] = new char[MAX_DEVSTRING_LENGTH];
		attr_SerialNumber_read[0] = new char[MAX_DEVSTRING_LENGTH];

		// Flush buffer
		try {
			_command("\n", rep);
		} catch(...) { /* Ingore any error */ }

		// Check device compatibility and retrieve firmware
		command("*IDN?\n", rep);

		stringstream ss(rep);
		string model, version;
		ss >> model;
		ss >> version;
		if (model != "PI-E816") {
			Tango::Except::throw_exception( "",
					model + " isn't supported",
					"E621::init_device()");
		}

		strncpy(attr_Firmware_read[0], version.c_str(), MAX_DEVSTRING_LENGTH);
		
		// Retrieve serial number
		command("SSN? A\n", rep);
		strncpy(attr_SerialNumber_read[0], rep.c_str(), MAX_DEVSTRING_LENGTH);

		check_state();
	} catch (Tango::DevFailed &e) {
		init_error = "Initialization failed: " + string(e.errors[0].desc);
	} catch(...) { 
		init_error = "Initialization failed: Unknown exception";
	}
	
	if(! init_error.empty()) {
		ERROR_STREAM << init_error << endl;
		set_state(Tango::UNKNOWN);
		set_status(init_error);
		assert(false);
	}
	
	/*----- PROTECTED REGION END -----*/	//	E621::init_device
}


//--------------------------------------------------------
/**
 *	Method     : E621::always_executed_hook()
 *	Description: method always executed before any command is executed
 */
//--------------------------------------------------------
void E621::always_executed_hook()
{
	DEBUG_STREAM << "E621::always_executed_hook()  " << device_name << std::endl;
	if (mandatoryNotDefined)
	{
		Tango::Except::throw_exception(
					(const char *)"PROPERTY_NOT_SET",
					get_status().c_str(),
					(const char *)"E621::always_executed_hook()");
	}
	/*----- PROTECTED REGION ID(E621::always_executed_hook) ENABLED START -----*/
	
	//	code always executed before all requests
	if(! init_error.empty()) {
		ERROR_STREAM << init_error << endl;
		set_state(Tango::UNKNOWN);
		set_status(init_error);
		assert(false);
		return;
	}

	string error;
	try {
		check_state();
	} catch (Tango::DevFailed &e) {
		error = string(e.errors[0].desc);
	} catch(...) { 
		error = "Unknown exception";
	}
	
	if (! error.empty()) {
		ERROR_STREAM << error << endl;
		set_state(Tango::UNKNOWN);
		set_status(error);
		assert(false);
	}

	/*----- PROTECTED REGION END -----*/	//	E621::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method     : E621::read_attr_hardware()
 *	Description: Hardware acquisition for attributes
 */
//--------------------------------------------------------
void E621::read_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "E621::read_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E621::read_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	check_init();

	/*----- PROTECTED REGION END -----*/	//	E621::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method     : E621::write_attr_hardware()
 *	Description: Hardware writing for attributes
 */
//--------------------------------------------------------
void E621::write_attr_hardware(TANGO_UNUSED(std::vector<long> &attr_list))
{
	DEBUG_STREAM << "E621::write_attr_hardware(std::vector<long> &attr_list) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E621::write_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	E621::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute Firmware related method
 *
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E621::read_Firmware(Tango::Attribute &attr)
{
	DEBUG_STREAM << "E621::read_Firmware(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E621::read_Firmware) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_Firmware_read);
	
	/*----- PROTECTED REGION END -----*/	//	E621::read_Firmware
}
//--------------------------------------------------------
/**
 *	Read attribute SerialNumber related method
 *
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E621::read_SerialNumber(Tango::Attribute &attr)
{
	DEBUG_STREAM << "E621::read_SerialNumber(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E621::read_SerialNumber) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_SerialNumber_read);
	
	/*----- PROTECTED REGION END -----*/	//	E621::read_SerialNumber
}
//--------------------------------------------------------
/**
 *	Read attribute AverageTimes related method
 *
 *
 *	Data type:	Tango::DevUShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E621::read_AverageTimes(Tango::Attribute &attr)
{
	DEBUG_STREAM << "E621::read_AverageTimes(Tango::Attribute &attr) entering... " << std::endl;
	/*----- PROTECTED REGION ID(E621::read_AverageTimes) ENABLED START -----*/
	//	Set the attribute value
	string rep;
	command("AVG?\n", rep);
	convert(rep, attr_AverageTimes_read[0]);
	attr.set_value(attr_AverageTimes_read);
	
	/*----- PROTECTED REGION END -----*/	//	E621::read_AverageTimes
}
//--------------------------------------------------------
/**
 *	Write attribute AverageTimes related method
 *
 *
 *	Data type:	Tango::DevUShort
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E621::write_AverageTimes(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "E621::write_AverageTimes(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevUShort	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(E621::write_AverageTimes) ENABLED START -----*/
	check_init();
	
	if (w_val != 1 && w_val != 2 && w_val != 4 && w_val != 8 && 
			w_val != 16 && w_val != 32 && w_val != 64) {
		Tango::Except::throw_exception( "",
				"Out of range (1,2,4,8,16,32,64)",
				"E621::write_AverageTimes()");
	}
		
 	stringstream cmd;
 	cmd << "AVG " << w_val << "\n";
 	command(cmd.str());

	/*----- PROTECTED REGION END -----*/	//	E621::write_AverageTimes
}

//--------------------------------------------------------
/**
 *	Read attribute Position related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E621::read_Position(Tango::Attribute &attr)
{
	DEBUG_STREAM << "E621::read_Position(Tango::Attribute &attr) entering... " << std::endl;
	Tango::DevDouble	*att_value = get_Position_data_ptr(attr.get_name());
	/*----- PROTECTED REGION ID(E621::read_Position) ENABLED START -----*/
	//	Set the attribute value
 	stringstream cmd;
 	cmd << "POS? " << attr.get_name().substr(attr.get_name().size()-1) << "\n";
 	string rep;
 	command(cmd.str(), rep);
 	convert(rep, *att_value);
	attr.set_value(att_value);
	
	/*----- PROTECTED REGION END -----*/	//	E621::read_Position
}
//--------------------------------------------------------
/**
 *	Write attribute Position related method
 *
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void E621::write_Position(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "E621::write_Position(Tango::WAttribute &attr) entering... " << std::endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(E621::write_Position) ENABLED START -----*/
	check_init();
	
 	stringstream cmd;
 	cmd << "MOV " << attr.get_name().substr(attr.get_name().size()-1) << " ";
 	cmd << fixed << setprecision(4) << setw(6) << setfill('0') << w_val << "\n";
 	command(cmd.str());
	
	/*----- PROTECTED REGION END -----*/	//	E621::write_Position
}
//--------------------------------------------------------
/**
 *	Method     : E621::add_dynamic_attributes()
 *	Description: Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void E621::add_dynamic_attributes()
{
	//	Example to add dynamic attribute:
	//	Copy inside the following protected area to create instance(s) at startup.
	//	add_Position_dynamic_attribute("MyPositionAttribute");

	/*----- PROTECTED REGION ID(E621::add_dynamic_attributes) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic attributes if any
	try {
		command("SAI?\n", sai_output);

		for (size_t i = 0; i < sai_output.size(); ++i) {
			/* Wave Table Output OFF */
			stringstream cmd;
			cmd << "WTO " << sai_output[i] << " 0\n";
			command(cmd.str());

			/* Servo-Control Mode ON (aka closed-loop) */
			/* SVO? does not report the hardware settings for the servo mode 
			 * but only the last sent SVO settings. (pag 13) */
			cmd.str(""); cmd.clear();
			cmd << "SVO " << sai_output[i] << " 1\n";
			command(cmd.str());

			/* Move Triggered Mode OFF */
			cmd.str(""); cmd.clear();
			cmd << "MVT " << sai_output[i] << " 0\n";
			command(cmd.str());

			/* Drift Compensation Mode ON */
			cmd.str(""); cmd.clear();
			cmd << "DCO " << sai_output[i] << " 1\n";
			command(cmd.str());


			stringstream name;
			name << "Position" << sai_output[i];
			add_Position_dynamic_attribute(name.str().c_str());
		}
	} catch (Tango::DevFailed &e) {
		init_error = string(e.errors[0].desc);
	} catch(...) { 
		init_error = "Unknown exception";
	}
	
	/*----- PROTECTED REGION END -----*/	//	E621::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Method     : E621::add_dynamic_commands()
 *	Description: Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void E621::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(E621::add_dynamic_commands) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic commands if any
	
	/*----- PROTECTED REGION END -----*/	//	E621::add_dynamic_commands
}

/*----- PROTECTED REGION ID(E621::namespace_ending) ENABLED START -----*/

//	Additional Methods
void E621::check_init()
{
	if(! init_error.empty())
		Tango::Except::throw_exception( "",
			init_error.c_str(),
			"E621::check_init()");
}

void E621::check_state()
{
	set_state(Tango::ON);
	set_status("The device is in ON state");

	/* I2C? and OVF? doesn't seem useful so use ONT? only */
	for (size_t i = 0; i < sai_output.size(); ++i) {
		stringstream cmd;
		cmd << "ONT? " << sai_output[i] << "\n";
		string rep;
		command(cmd.str(), rep);

		int code;
		convert(rep, code);
		if (code != 1) {
			set_state(Tango::MOVING);
			set_status("The device is in MOVING state");
			return;
		}
	}
}

void E621::_command(const string &cmd)
{
	/* More than one command mnemonic per line is not allowed. (pag 34) */
 	/* Only one axis can be addressed per command. (pag 34) */

	/* The number of characters per line is limited to 25 bytes. (pag 34) */
	assert(cmd.size() < 26);

	write(cmd);
}

void E621::command(const string &cmd)
{
	_command(cmd);
	verify_command_execution();
}

bool invalidChar (char c) 
{  
    return !(c>=32 && c <126);  
} 

void E621::_command(const string &cmd, string &rep)
{
	_command(cmd);
	read_until('\n', rep);

	/* When controlling the E-816, timing problems can occur if several commands
	 * are run in rapid sequence, resulting in lost commands. To prevent such 
	 * communication errors, it is recommended that you include a certain wait time
	 * between the different commands, depending on the command to be executed. 
	 * This is especially true for commands that need a certain execution time 
	 * inside the E-816 module, like MOV, MVR, SPA, SVA, SVR, RST, WPA, SWT 
	 * and WTO. (pag 24, 34) */
    	rep.erase(remove_if(rep.begin(),rep.end(), invalidChar), rep.end());  
}

void E621::command(const string &cmd, string &rep)
{
	bool check_error = false;

	try {
		_command(cmd, rep);
	} catch (...) {
		check_error = true;
	}

	if (check_error)
		verify_command_execution();
}

void E621::verify_command_execution()
{
	string rep;
	_command("ERR?\n", rep);

	int errcode;
	convert(rep, errcode);

	string errmesg;
	switch(errcode) {
		case 1: errmesg = "Parameter syntax error"; break;
		case 2: errmesg = "Unknown command"; break;
		case 3: errmesg = "Command length out of limits or command "
				"buffer overrun"; break;
		case 5: errmesg = "Unallowable move attempted on unreferenced "
				"axis, or move attempted with servo off"; break;
		case 10: errmesg = "Controller was stopped by command"; break;
		case 15: errmesg = "Invalid axis identifier"; break;
		case 17: errmesg = "Parameter out of range"; break;
		case 20: errmesg = "Macro not found"; break;
		case 54: errmesg = "Unknown parameter"; break;
		case 56: errmesg = "Password invalid"; break;
		case 60: errmesg = "Protected Param: current Command Level (CCL)"
			 	" too low"; break;
		case 73: errmesg = "Motion commands are not allowed when wave "
			 	"generator is  active"; break;
		case 79: errmesg = "Open-loop commands (SVA, SVR) are not "
			 	"allowed when servo is on"; break;
		case 89: errmesg = "Command not allowed in current motion mode"; break;
		case 210: errmesg = "Illegal file name (must be 8-0 format)"; break;
		case 232: errmesg = "Save system configuration failed"; break;
		case 233: errmesg = "Load system configuration failed"; break;
		case 306: errmesg = "Error on I2C bus";	break;
		case 309: errmesg = "Insufficient space to store macro"; break;
		case 405: errmesg = "Wave parameter out of range"; break;
	}

	if (! errmesg.empty()) {
		ERROR_STREAM << errmesg << endl;
		assert(false);
		Tango::Except::throw_exception( "",
				"Command execution failed: " + errmesg,
				"E621::verify_command_execution()");
	}
}

/*----- PROTECTED REGION END -----*/	//	E621::namespace_ending
} //	namespace
